from typing import Union, Sequence
from functools import lru_cache

from werkzeug.utils import cached_property
import numpy as np
from scipy.spatial.transform import Rotation

L = Union[Sequence, np.ndarray]


def _immutable(a) -> np.ndarray:
    if isinstance(a, np.ndarray):
        if a.flags.writeable:
            a = a.copy()
            a.setflags(write=False)
    else:
        a = np.array(a)
        a.setflags(write=False)
    return a


_R_ident = _immutable(np.eye(3))
_r_ident = Rotation.identity()
_t_ident = _immutable(np.zeros(3))


class Transform:
    def __init__(self, matrix: L = None, R: L = None, t: L = None, r: Rotation = None,
                 quat: L = None, euler: L = None, rpy: L = None, rotvec: L = None, inv=None, degrees=False,
                 validate=True):
        self._r = None
        self._inv = inv  # type: Union[None, Transform]
        if matrix is not None:
            self.matrix = _immutable(matrix)
            self.R = matrix[:3, :3]
            self.t = matrix[:3, 3]
        else:
            if R is not None:
                self.R = _immutable(R)
            else:
                # if rotation is not provided as a matrix, convert using scipy.spatial.Rotation
                if quat is not None:
                    r = Rotation.from_quat(quat)
                elif rotvec is not None:
                    r = Rotation.from_rotvec(np.deg2rad(rotvec) if degrees else rotvec)
                elif euler is not None:
                    seq, angles = euler
                    r = Rotation.from_euler(seq, angles, degrees=degrees)
                elif rpy is not None:
                    r = Rotation.from_euler('xyz', rpy, degrees=degrees)

                self._r = r
                if r is None:  # no rotation provided
                    self.R = _R_ident
                else:
                    self.R = self._r.as_matrix()
                    self.R.setflags(write=False)
            self.t = _t_ident if t is None else _immutable(t)
            self.matrix = np.eye(4)
            self.matrix[:3, :3] = self.R
            self.matrix[:3, 3] = self.t
            self.matrix.setflags(write=False)

    @property
    def inv(self):
        if self._inv is None:
            self._inv = Transform(
                R=self.R.T,
                t=-self.t @ self.R,
                inv=self
            )
        return self._inv

    @property
    def r(self) -> Rotation:
        if self._r is None:
            self._r = Rotation.from_matrix(self.R)
        return self._r

    @cached_property
    def quat(self):
        return self.r.as_quat()

    @cached_property
    def rotvec(self):
        return self.r.as_rotvec()

    @lru_cache()
    def euler(self, seq):
        return self.r.as_euler(seq)

    @property
    def rpy(self):
        return self.euler('xyz')

    @property
    def xyz_rotvec(self):
        return (*self.t, *self.rotvec)

    @classmethod
    def from_xyz_rotvec(cls, xyz_rotvec):
        return cls(t=xyz_rotvec[:3], rotvec=xyz_rotvec[3:])

    def __matmul__(self, other):
        if isinstance(other, Transform):
            matrix = self.matrix @ other.matrix
            matrix.setflags(write=False)
            return Transform(matrix=matrix)
        elif isinstance(other, (np.ndarray, Sequence)):
            return np.asarray(other) @ self.R.T + self.t
        else:
            raise NotImplementedError()

    def rotate(self, other):
        other = np.asarray(other)
        d = other.shape[-1]
        if d == 6:  # screw
            return np.concatenate((
                self.rotate(other[..., :3]),
                self.rotate(other[..., 3:])
            ))
        assert d == 3
        return other @ self.R.T

    def __mul__(self, scale):
        return Transform(
            t=self.t * scale,
            rotvec=self.rotvec * scale,
        )

    def lerp(self, other: 'Transform', s: float):
        self_t_other = self.inv @ other
        return self @ (self_t_other * s)

    def save(self, fp):
        np.savetxt(fp, self.xyz_rotvec)

    @classmethod
    def load(cls, fp):
        cls.from_xyz_rotvec(np.loadtxt(fp))

    @classmethod
    def random(cls, random_state: np.random.RandomState = np.random):  # for testing
        return cls(t=random_state.randn(3), r=Rotation.random(random_state=random_state))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(self.xyz_rotvec)

    def equals(self, other, atol=1e-13):
        err = np.abs(self.matrix - other.matrix)
        return np.all(err < atol)



import numpy as np

from .. import Transform


def test_transform_operations(atol=1e-13):
    for _ in range(100):
        t = np.random.randn(3)
        rpy = np.random.randn(3)
        f = Transform(t=t, rpy=rpy)

        # matmul
        f_ = Transform(t=t) @ Transform(rpy=rpy)
        err = np.abs(f.matrix - f_.matrix).max()
        assert err < atol, err

        # inverse
        assert np.abs((f.inv @ f).matrix - np.eye(4)).max() < atol
        assert np.abs(np.linalg.inv(f.matrix) - f.inv.matrix).max() < atol


def test_point_identity_transformation():
    t = Transform()
    assert np.all(np.ones(3) == t @ np.ones(3))
    p = np.random.randn(100, 3)
    assert np.all(p == t @ p)



import re
from typing import Union, Dict, List, Sequence

from .transform import Transform

T = Union[Transform, str]


class SceneState:
    def __init__(self, transforms=None):
        if transforms is None:
            transforms = dict()  # type: Dict[SceneNode, Transform]
        self.transforms = transforms

    def __getitem__(self, node: 'SceneNode'):
        return self.transforms[node]

    def __setitem__(self, node: Union['SceneNode', Sequence['SceneNode']],
                    transform: Union[T, Sequence[T]]):
        nodes = node if isinstance(node, Sequence) else [node]
        transforms = transform if isinstance(transform, Sequence) else [transform]
        transforms = [Transform.load(t) if isinstance(t, str) else t for t in transforms]
        if len(transforms) == 1:
            transforms *= len(nodes)
        assert len(transforms) == len(nodes)
        for n, t in zip(nodes, transforms):
            self.transforms[n] = t

    def T(self, a: 'SceneNode', b: 'SceneNode'):
        return a.t(b, self)

    def copy(self):
        return SceneState(self.transforms.copy())


class SceneNode:
    def __init__(self, parent: 'SceneNode' = None, name='SceneNode'):
        self.parent = parent
        self.name = name

    @classmethod
    def tree(cls, s: str):
        """
        Example:
            root, a, b, c = SceneNode.from_str('root -> (a, b -> c)')
        """
        s = re.sub(r'\s', '', s) + ','
        nodes = []
        last_b = 0
        stack = [None]
        p_stack = [None]
        for match in re.finditer(r'(->|[,()])', s):
            a, b = match.span()
            m = s[a:b]
            if m in ('->', ',', ')') and a - last_b:
                name = s[last_b:a]
                assert name
                nodes.append(cls(parent=stack[-1], name=name))
            if m == '->':
                stack.append(nodes[-1])
            if m == '(':
                p_stack.append(stack[-1])
            if m == ')':
                p_stack.pop()
                m = ','
            if m == ',':
                while stack[-1] != p_stack[-1]:
                    stack.pop()
            last_b = b
        return tuple(nodes)

    def path_to_root(self):
        nodes = [self]
        while nodes[-1].parent:
            nodes.append(nodes[-1].parent)
        return nodes

    def path_to(self, other: 'SceneNode'):
        root_to_self = self.path_to_root()[::-1]
        root_to_other = other.path_to_root()[::-1]
        idx = ([a is b for a, b in zip(root_to_self, root_to_other)] + [False]).index(False) - 1
        assert idx >= 0
        self_to_common_parent = root_to_self[idx:][::-1]
        common_parent_to_other = root_to_other[idx:]
        return self_to_common_parent, common_parent_to_other

    def common_parent(self, other):
        return self.path_to(other)[0][-1]

    def t(self, other: 'SceneNode', state: SceneState):
        """
        Returns other in self's frame
        """
        self_to_common, common_to_other = self.path_to(other)
        t = Transform()
        for node in self_to_common[:-1]:
            t = t @ state[node]
        for node in common_to_other[1:]:
            t = t @ state[node].inv
        return t

    def solve(self, a: 'SceneNode', b: 'SceneNode', a_t_b: Transform, state: SceneState):
        """
        Returns the transform that should be applied to self to achieve a_T_b.
        """
        a_to_common, common_to_b = a.path_to(b)
        b_t_a = a_t_b.inv
        if self in a_to_common[:-1]:
            return self.t(a, state) @ a_t_b @ b.t(self.parent, state)
        elif self in common_to_b[1:]:
            return self.t(b, state) @ b_t_a @ a.t(self.parent, state)
        else:
            ValueError('self.parent must be in the path from a to b')

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)



import numpy as np

from .. import Transform, SceneNode, SceneState


def test_from_str():
    root, a, b, c, d, e = SceneNode.tree('root -> (a, b -> c, d), e')
    for node, parent in (root, None), (a, root), (b, root), (c, b), (d, root), (e, None):
        assert node.parent is parent


def test_t(atol=1e-13):
    state = SceneState()
    root = node = SceneNode()
    for _ in range(5):
        node = SceneNode(node)
        state[node] = Transform.random()

    child = node
    mat = np.eye(4)
    while node != root:
        mat = mat @ state[node].matrix
        node = node.parent
    assert np.abs(mat - child.t(root, state).matrix).max() < atol


def test_path_to_and_solve(atol=1e-13):
    root, base, a, c, b = SceneNode.tree('root -> base -> (a, c -> b)')
    assert a.path_to(root) == ([a, base, root], [root])
    assert a.path_to(b) == ([a, base], [base, c, b])
    assert c.path_to(b) == ([c], [c, b])

    state = SceneState()
    state[a, c, b] = [Transform.random() for _ in range(3)]

    desired_a_t_b = Transform.random()
    state[c] = c.solve(a, b, desired_a_t_b, state)
    err = np.abs(desired_a_t_b.matrix - a.t(b, state).matrix).max()
    assert err < atol, err



import setuptools

with open("readme.md") as f:
    long_description = f.read()

setuptools.setup(
    name='transform3d',
    version='0.0.0',
    author='Rasmus Laurvig Haugaard',
    author_email='rasmus.l.haugaard@gmail.com',
    description='handy classes for 3d transformations',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/RasmusHaugaard/transform3d',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.18',
        'scipy>=1.4',
        'werkzeug>=0.16'
    ],
    python_requires='>=3.6',
)